#
# FtpCube
# Copyright (C) 2002 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import main
import messages
import events
import dispatcher
import browser
import logger
import cache
import ctrlwin
import chmodwin
import util
import protocol
import icons.folder
import icons.link
import icons.file

import os
import re
import time
from wxPython.wx import *

class AbstractFileWindow (wxPanel):

	UNSORTED = 0
	SORT_BY_NAME = 1
	SORT_BY_SIZE = 2
	SORT_BY_DATE = 3

	idLIST = wxNewId ()

	hidden_re = re.compile ("^\..*")

	def __init__ (self, parent, headers):
		wxPanel.__init__ (self, parent, -1)

		self.list = wxListCtrl (self, self.idLIST, style=wxLC_REPORT | wxSUNKEN_BORDER)
		for i, name in zip (range (len (headers)), headers):
			self.list.InsertColumn (i, name)

		self.image_list = wxImageList (15, 17)
		self.folder_index = self.image_list.Add (icons.folder.getBitmap ())
		self.link_index = self.image_list.Add (icons.link.getBitmap ())
		self.file_index = self.image_list.Add (icons.file.getBitmap ())
		self.list.SetImageList (self.image_list, wxIMAGE_LIST_SMALL)

		status_bar = self.makeStatusBar ()

		sizer = wxBoxSizer (wxVERTICAL)
		sizer.Add (self.list, 1, wxEXPAND)
		sizer.Add (status_bar, 0, wxEXPAND)
		self.SetAutoLayout (true)
		self.SetSizer (sizer)

		# Set instance variables
		self.sorted = self.SORT_BY_NAME
		self.hidden_files = 1
		self.directories_first = 1
		self.dir = None

		# Set up the interactivity
		EVT_LEFT_DCLICK (self.list, self.onDoubleClick)
		EVT_RIGHT_DOWN (self.list, self.onRightClick)

	def makeStatusBar (self):
		self.status_label = wxStaticText (self, -1, "")
		hsizer = wxBoxSizer (wxHORIZONTAL)
		hsizer.Add (5, 1)
		hsizer.Add (self.status_label, 1, wxEXPAND)
		vsizer = wxBoxSizer (wxVERTICAL)
		vsizer.Add (0, 3)
		vsizer.Add (hsizer, 0, wxEXPAND)
		vsizer.Add (0, 3)
		return vsizer

	def getStatusBar (self):
		return self.status_label

	def updateStatusBar (self, dir):
		self.getStatusBar ().SetLabel (dir)

	def selectBitmapIndex (self, flags):
		if not flags:
			return self.folder_index
		else:
			if flags[0] == 'l':
				return self.link_index
			elif flags[0] in  'd ':
				return self.folder_index
			else:
				return self.file_index

	def setDir (self, dir):
		self.dir = dir

	def getDir (self):
		return self.dir

	def onDoubleClick (self, event):
		raise NotImplementedError

	def onRightClick (self, event):
		raise NotImplementedError

	def toggleHide (self, event=None):
		self.hidden_files = (self.hidden_files + 1) % 2
		self.updateListing ('.')

	def toggleDirectoriesFirst (self, event=None):
		self.directories_first = (self.directories_first + 1) % 2
		self.updateListing ('.')

	def updateListing (self, dir):
		raise NotImplementedError

	def unsorted (self, event=None):
		self.sorted = self.UNSORTED
		self.directories_first = 0
		self.updateListing ('.')

	def sortByName (self, event=None):
		self.sorted = self.SORT_BY_NAME
		self.directories_first = 1
		self.updateListing ('.')

	def sortBySize (self, event=None):
		self.sorted = self.SORT_BY_SIZE
		self.directories_first = 1
		self.updateListing ('.')

	def sortByDate (self, event=None):
		self.sorted = self.SORT_BY_DATE
		self.directories_first = 1
		self.updateListing ('.')

	def clearList (self):
		if self.list:
			self.list.Freeze ()
			self.list.DeleteAllItems ()
			self.list.Thaw ()

	def sortList (self, sequence):
		raise NotImplementedError

	def performSort (self, sequence):
		if self.sorted == self.SORT_BY_NAME:
			sequence.sort (lambda x, y: cmp (x[0], y[0]))
		elif self.sorted == self.SORT_BY_SIZE:
			sequence.sort (lambda x, y: cmp (long (x[4]), long (y[4])))
		elif self.sorted == self.SORT_BY_DATE:
			# Convert the times beforehand for speed
			try:
				converted = [ (i, time.strptime (i[2], "%b %d %H:%M")) for i in sequence ]
			except Exception:
				return sequence
			converted.sort (lambda x, y: cmp (x[1], y[1]))
			sequence = [ i[0] for i in converted ]
		return sequence

	def getFileSize (self, file):
		index = self.list.FindItem (-1, file)
		return util.getColumnText (self.list, index, 2)

	def getMainThread (self):
		thread_mgr = main.getApp ().getThreadManager ()
		return thread_mgr.getMainThread ()

class LocalWindow (AbstractFileWindow):

	idPOPUP_UPLOAD  = wxNewId ()
	idPOPUP_RENAME  = wxNewId ()
	idPOPUP_DELETE  = wxNewId ()
	idPOPUP_CREATE  = wxNewId ()
	idPOPUP_CHMOD   = wxNewId ()
	idPOPUP_CHANGE  = wxNewId ()
	idPOPUP_REFRESH = wxNewId ()
	idPOPUP_HIDDEN  = wxNewId ()
	idPOPUP_UNSORT  = wxNewId ()
	idPOPUP_BYNAME  = wxNewId ()
	idPOPUP_BYSIZE  = wxNewId ()
	idPOPUP_BYDATE  = wxNewId ()
	idPOPUP_ORDER   = wxNewId ()

	time_re = re.compile ('[A-Za-z]+\s+([A-Za-z]+\s+\d+\s+\d+:\d+):(\d+)\s+\d+')

	def __init__ (self, parent):
		if __debug__:
			print "Making local file browsing window."

		self.headers = [ _("Filename"), _("Size"), _("Date"), _("Flags") ]
		AbstractFileWindow.__init__ (self, parent, self.headers)

		# Set the date alignment
		util.setListColumnAlignment (self.list, 2, wxLIST_FORMAT_RIGHT)

		# Set the headings sizes accordingly
		for i in range (len (self.headers)):
			self.list.SetColumnWidth (i, 100)
		# Explicitly make the filenames column bigger
		self.list.SetColumnWidth (0, 200)

		# Open up the default directory for display
		if sys.platform == 'win32':
			if os.environ.has_key ('HOMEPATH'):
				self.setDir (os.environ['HOMEPATH'])
			else:
				self.setDir (os.getcwd ())
		elif os.name == 'posix':
			if os.environ.has_key ('HOME'):
				self.setDir (os.environ['HOME'])
			else:
				self.setDir (os.getcwd ())
		else:
			self.setDir (os.getcwd ())

		self.last_listing = None
		self.updateListing (self.getDir ())

		events.EVT_LOCAL_WINDOW (self, self.onLocalWindowEvent)
		main.getApp ().registerEventListener (events.wxEVT_LOCAL_WINDOW, self)

	def onDoubleClick (self, event):
		item = self.list.GetNextItem (-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED)
		if item != -1:
			selected = util.getColumnText (self.list, item, 0)
			self.updateListing (selected)

	def onRightClick (self, event):
		menu = self.makePopupMenu ()
		self.PopupMenu (menu, wxPoint (event.m_x, event.m_y))

	def onLocalWindowEvent (self, event):
		if event.kind == events.wxLocalWindowEvent.EVT_LIST_UPDATE:
			self.updateListing (event.msg)
		elif event.kind == events.wxLocalWindowEvent.EVT_LIST_REFRESH:
			self.onRefresh (event)

	def makePopupMenu (self):
		# Create the popup menu
		menu = wxMenu ()
		menu.Append (self.idPOPUP_UPLOAD, _("Upload Files"))
		menu.Append (self.idPOPUP_RENAME, _("Rename Files"))
		menu.Append (self.idPOPUP_DELETE, _("Delete Selected Files"))
		menu.AppendSeparator ()
		menu.Append (self.idPOPUP_CREATE, _("Create Directory"))
		menu.Append (self.idPOPUP_CHMOD, _("Change Permissions"))
		menu.Append (self.idPOPUP_CHANGE, _("Change Directory"))
		menu.Append (self.idPOPUP_REFRESH, _("Refresh Listing"))
		menu.AppendSeparator ()
		menu.AppendCheckItem (self.idPOPUP_HIDDEN, _("Hide Hidden Files"))
		menu.AppendSeparator ()

		# Create the sort order sub menu and add it
		sort_menu = wxMenu ()
		sort_menu.Append (self.idPOPUP_UNSORT, _("Unsorted"))
		sort_menu.Append (self.idPOPUP_BYNAME, _("Sort by Name"))
		sort_menu.Append (self.idPOPUP_BYSIZE, _("Sort by Size"))
		sort_menu.Append (self.idPOPUP_BYDATE, _("Sort by Date"))
		sort_menu.AppendSeparator ()
		sort_menu.AppendCheckItem (self.idPOPUP_ORDER, _("Order Directories-Links-Files"))
		menu.AppendMenu (wxNewId (), _("Sort Order"), sort_menu)

		if self.hidden_files:
			menu_item = menu.FindItemById (self.idPOPUP_HIDDEN)
			menu_item.Check (true)

		if self.directories_first:
			menu_item = menu.FindItemById (self.idPOPUP_ORDER)
			menu_item.Check (true)

		# Set the event handlers
		EVT_MENU (self, self.idPOPUP_UPLOAD, self.onUpload)
		EVT_MENU (self, self.idPOPUP_RENAME, self.onRename)
		EVT_MENU (self, self.idPOPUP_DELETE, self.onDelete)
		EVT_MENU (self, self.idPOPUP_CREATE, self.onCreate)
		EVT_MENU (self, self.idPOPUP_CHMOD, self.onChmod)
		EVT_MENU (self, self.idPOPUP_CHANGE, self.onChange)
		EVT_MENU (self, self.idPOPUP_REFRESH, self.onRefresh)
		EVT_MENU (self, self.idPOPUP_HIDDEN, self.toggleHide)
		EVT_MENU (self, self.idPOPUP_UNSORT, self.unsorted)
		EVT_MENU (self, self.idPOPUP_BYNAME, self.sortByName)
		EVT_MENU (self, self.idPOPUP_BYSIZE, self.sortBySize)
		EVT_MENU (self, self.idPOPUP_BYDATE, self.sortByDate)
		EVT_MENU (self, self.idPOPUP_ORDER, self.toggleDirectoriesFirst)

		return menu

	def onUpload (self, event):
		selected = util.getSelected (self.list)
		main_thread = main.getApp ().getThreadManager ().getMainThread ()
		opts = main_thread.getOpts ()
		remote_win = main.getApp ().getMainWindow ().getRemoteWindow ()
		for item in selected:
			file = util.getColumnText (self.list, item, 0)
			if file == '..':
				continue
			path = os.path.join (self.getDir (), file)
			flags = self.last_listing[self.list.GetItemData (item)][3]

			# Check if the file exists remotely. If so replace
			# the file's size with the remote size. Later we'll
			# check if the size is smaller and perform a resume
			# if it is
			remote_size = remote_win.getTrueFileSize (file)
			if remote_size is None:
				try:
					remote_size = os.path.getsize (path)
				except OSError:
					remote_size = 0

			event = events.wxQueueEvent (
				file        =  file,
				host        = opts['host'],
				port        = opts['port'],
				username    = opts['username'],
				password    = opts['password'],
				size        = remote_size,
				flags       = flags[0], # Just provide the type flag
				remote_path = remote_win.getDir (),
				local_path  = self.getDir (),
				direction   = protocol.UPLOAD,
				attempt     = 1,
				maxattempts = opts['retries'],
				method      = main.getApp ().getTransferMethod (),
			)
			main.getApp ().postEvent (event)

	def onRename (self, event):
		selected = util.getSelected (self.list)
		for item in selected:
			file = util.getColumnText (self.list, item, 0)
			orig_path = os.path.join (self.getDir (), file)
			new = messages.displayInputDialog (self, _("Ftpcube - Rename"),
				_("Rename %(f)s to:") %{ 'f' : file })
			if new:
				new_path = os.path.join (self.getDir (), new)
				try:
					os.rename (orig_path, new_path)
				except OSError, strerror:
					messages.displayErrorDialog (
						self, _("Error renaming file %(orig)s to %(new)s: %(err)s")
						%{ 'orig' : orig_path, 'new' : new_path, 'err' : strerror }
					)
					return
		self.updateListing (self.getDir ())

	def onDelete (self, event):
		selected = util.getSelected (self.list)
		for item in selected:
			file = util.getColumnText (self.list, item, 0)
			type = self.last_listing[self.list.GetItemData (item)][3]
			path = os.path.join (self.getDir (), file)
			if type[0] in '-lf':
				try:
					os.remove (path)
				except OSError, strerror:
					messages.displayErrorDialog (
						self, _("Error removing file: ") + strerror
					)
			else:
				try:
					os.rmdir (path)
				except OSError, strerror:
					messages.displayErrorDialog (
						self, _("Error removing directory: " ) + strerror
					)
			self.list.SetItemState (item, 0, wxLIST_STATE_SELECTED)
		self.updateListing (self.getDir ())

	def onCreate (self, event):
		dir = messages.displayInputDialog (self, _("Ftpcube - Create Directory"),
		                                         _("Directory Name:"))
		if dir:
			path = os.path.join (self.getDir (), dir)
			try:
				os.mkdir (path)
			except OSError, strerror:
				messages.displayErrorDialog (
					self, _("Error creating directory: ") + strerror
				)
				return
			self.updateListing (self.getDir ())

	def onChmod (self, event):
		selected = util.getSelected (self.list)
		for item in selected:
			file = util.getColumnText (self.list, item, 0)
			path = os.path.join (self.getDir (), file)
			try:
				info = os.stat (path)
				perm = info[0] # st_mode
			except OSError, strerror:
				messages.displayErrorDialog (self,
					_("Error obtaining permissions for %(f)s: %(err)s")
					%{ 'f' : file, 'err' : strerror })
				continue

			chmod_win = chmodwin.ChmodWindow (self, file, perm)
			ret = chmod_win.ShowModal ()
			if ret == wxID_OK:
				new_perm = chmod_win.getPermissions ()
				try:
					os.chmod (path, new_perm)
				except OSError, strerror:
					messages.displayErrorDialog (self,
						_("Error setting permissions for %(f)s: %(err)s")
						%{ 'f' : file, 'err' : strerror })
		if selected:
			self.updateListing (self.getDir ())

	def onChange (self, event):
		local_browser = browser.LocalBrowser (self, self.getDir ())
		ret = local_browser.ShowModal ()
		if ret == wxID_OK:
			dir = local_browser.getDirectory ()
			if dir:
				self.updateListing (dir)

	def onRefresh (self, event):
		self.updateListing ('.')

	def updateListing (self, dir):
		dir = os.path.normpath (dir)

		# Construct our new path
		if dir == '..':
			self.setDir (os.path.dirname (self.getDir ()))
		elif dir != '.':
			newpath = os.path.join (self.getDir (), dir)
			if os.path.isdir (newpath):
				self.setDir (newpath)
			else:
				messages.displayErrorDialog (self, _("Invalid directory: ") + newpath)

		# Get the new listing
		self.last_listing = self.readDir (self.getDir ())

		# Now change the list to reflect the new listing. Notice that
		# we pause the visuals so it doesn't look funny
		self.list.Freeze ()
		self.clearList ()
		if self.sorted:
			parent_entry = self.last_listing.pop (0)
			self.last_listing = self.sortList (self.last_listing)
			# Make sure parent entry is first
			self.last_listing.insert (0, parent_entry)
		for i in range (len (self.last_listing)):
			item = self.last_listing[i]
			img_index = self.selectBitmapIndex (item[3])
			self.list.InsertImageStringItem (i, item[0], img_index)
			self.list.SetStringItem (i, 1, item[1])
			self.list.SetStringItem (i, 2, item[2])
			self.list.SetStringItem (i, 3, item[3])
			self.list.SetItemData (i, i)
		self.list.Thaw ()

		self.updateStatusBar (self.getDir ())

	def readDir (self, dir):
		try:
			files = os.listdir (dir)
		except OSError, strerror:
			messages.displayErrorDialog (self, strerror)
			return

		if self.hidden_files:
			files = [ i for i in files if not self.hidden_re.match (i) ]

		# Add the special file ".."
		files.insert (0, '..')

		listitems = [ ]
		for f in files:
			# Attempt to retrieve the file info and modification times
			try:
				# lstat is preferrable for systems that support it
				if os.name == 'posix':
					(st_mode, st_ino, st_dev, st_nlink, st_uid, st_gid, st_size,
					 st_atime, st_mtime, st_ctime) = os.lstat (os.path.join (dir, f))
				else:
					(st_mode, st_ino, st_dev, st_nlink, st_uid, st_gid, st_size,
					 st_atime, st_mtime, st_ctime) = os.stat (os.path.join (dir, f))
			except OSError, strerror:
				# Set the size and mtime to 0 then and carry on if we can't stat
				# the file for some reason but know it's there
				st_size = 0
				st_mtime = 0

			# Beautify the size string
			size = util.beautifySize (st_size)

			# Beautify the time
			match = self.time_re.match (time.ctime (st_mtime))
			mtime = match.group (1)

			# Set the appropriate file type and determine the mode
			mode = self.convertPermToStr (st_mode)
			if os.path.islink (os.path.join (self.getDir (), f)):
				mode = "l" + mode
			elif os.path.isdir (os.path.join (self.getDir (), f)):
				mode = "d" + mode
			else:
				mode = "-" + mode

			listitems.append ((f, size, mtime, mode, str (st_size)))

		return listitems

	def convertPermToStr (self, perm):
		flags = [ '-' ] * 9
		if perm & (4 << 6):
			flags[0] = 'r'
		if perm & (2 << 6):
			flags[1] = 'w'
		if perm & (1 << 6):
			flags[2] = 'x'
		if perm & (4 << 3):
			flags[3] = 'r'
		if perm & (2 << 3):
			flags[4] = 'w'
		if perm & (1 << 3):
			flags[5] = 'x'
		if perm & 4:
			flags[6] = 'r'
		if perm & 2:
			flags[7] = 'w'
		if perm & 1:
			flags[8] = 'x'
		return ''.join (flags)

	def sortList (self, list):
		if self.directories_first:
			dirs = [ i for i in list if i[3][0] == 'd' ]
			links = [ i for i in list if i[3][0] == 'l' ]
			files = [ i for i in list if i[3][0] in '-f' ]

			list = [ ]
			list.extend (self.performSort (dirs))
			list.extend (self.performSort (links))
			list.extend (self.performSort (files))
		else:
			list = self.performSort (list)
		return list

class RemoteWindow (AbstractFileWindow):

	idPOPUP_DOWNLOAD = wxNewId ()
	idPOPUP_RENAME   = wxNewId ()
	idPOPUP_DELETE   = wxNewId ()
	idPOPUP_CREATE   = wxNewId ()
	idPOPUP_REMOVE   = wxNewId ()
	idPOPUP_CHANGE   = wxNewId ()
	idPOPUP_CHMOD    = wxNewId ()
	idPOPUP_REFRESH  = wxNewId ()

	idPOPUP_UNSORT     = wxNewId ()
	idPOPUP_SORTBYNAME = wxNewId ()
	idPOPUP_SORTBYSIZE = wxNewId ()
	idPOPUP_SORTBYDATE = wxNewId ()
	idPOPUP_ORDER      = wxNewId ()

	list_re = re.compile ('([A-Za-z-]+)\s+(\d+)\s+([A-Za-z0-9]+)\s+([A-Za-z0-9]+)\s+(\d+)\s+([A-Za-z]+\s+[0-9]+\s+[0-9:]+)\s+(.*)', re.IGNORECASE)
	ms_list_re = re.compile ('([A-Za-z-]+)\s+(\d+)\s+([A-Za-z]+)\s+(\d+)\s+([A-Za-z]+\s+\d+\s+[0-9:]+)\s+(.*)')
	link_re = re.compile ('(.*) -> (.*)')
	broken_link_re = re.compile ('.*-> (.*)')
	time_re = re.compile ('[A-Za-z]+\s+([A-Za-z]+\s+\d+\s+\d+:\d+):(\d+)\s+\d+')

	def __init__ (self, parent):
		if __debug__:
			print "Making remote file browsing window."

		self.headers = [ _("Filename"), _("Size"), _("Date"), _("Flags") ]
		AbstractFileWindow.__init__ (self, parent, self.headers)

		# Set the date alignment
		util.setListColumnAlignment (self.list, 2, wxLIST_FORMAT_RIGHT)
		# Set the headings sizes accordingly
		for i in range (len (self.headers)):
			self.list.SetColumnWidth (i, 100)
		# Explicitly make the filenames column bigger
		self.list.SetColumnWidth (0, 200)

		self.cache = cache.CacheTree ()

		events.EVT_REMOTE_WINDOW (self, self.onRemoteWindowEvent)
		main.getApp ().registerEventListener (events.wxEVT_REMOTE_WINDOW, self)

	def onDoubleClick (self, event):
		item = self.list.GetNextItem (-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED)
		if item != -1:
			selected = util.getColumnText (self.list, item, 0)

			thread = self.getMainThread ()
			if selected == '..':
				old_dir = self.getDir ()
				mydir = self.dirname (old_dir)
				clogger = logger.ConsoleLogger ()
				handler = CwdErrorHandler (old_dir, self, clogger)
				thread.cwd ('..', handler=handler)
				self.updateStatusBar (mydir)
				self.setDir (mydir)
				self.getListing ()
			else:
				type = util.getColumnText (self.list, item, 3)
				if type[0] in 'ld':
					old_dir = self.getDir ()
					mydir = self.join ([ old_dir, selected ])
					self.changeDirectory (mydir, old_dir)

	def dirname(self, dir):
		index = dir.rfind('/')
		if index == 0:
			return '/'
		else:
			return dir[:index]

	def normpath(self, path):
		dirs = path.split('/')
		pathlist = [ ]
		i = 0
		while i < len (dirs):
			if dirs[i] == '.':
				pass
			elif dirs[i] == '..':
				i = i + 1
			elif dirs[i]:
				pathlist.append (dirs[i])
			i = i + 1
		return '/' + '/'.join (pathlist)

	def join(self, args):
		return self.normpath ('/'.join (args))

	def changeDirectory (self, dir, old_dir):
		main_thread = self.getMainThread ()
		if main_thread is not None:
			console_logger = logger.ConsoleLogger ()
			handler = CwdErrorHandler (old_dir, self, console_logger)
			self.updateStatusBar (dir)
			self.setDir (dir)
			main_thread.cwd (dir, handler=handler)
			self.getListing ()

	def onRightClick (self, event):
		menu = self.makePopupMenu ()
		self.PopupMenu (menu, wxPoint (event.m_x, event.m_y))

	def makePopupMenu (self):
		menu = wxMenu ()
		menu.Append (self.idPOPUP_DOWNLOAD, _("Download Selected"))
		menu.Append (self.idPOPUP_RENAME, _("Rename Selected"))
		menu.Append (self.idPOPUP_CHMOD, _("Change Permissions"))
		menu.Append (self.idPOPUP_DELETE, _("Delete Selected"))
		menu.AppendSeparator ()
		menu.Append (self.idPOPUP_CREATE, _("Create Directory"))
		menu.Append (self.idPOPUP_REMOVE, _("Remove Directory"))
		menu.Append (self.idPOPUP_CHANGE, _("Change Directory"))
		menu.Append (self.idPOPUP_REFRESH, _("Refresh Cache"))
		menu.AppendSeparator ()

		sort_menu = wxMenu ()
		sort_menu.Append (self.idPOPUP_UNSORT, _("Unsorted"))
		sort_menu.Append (self.idPOPUP_SORTBYNAME, _("Sort by Name"))
		sort_menu.Append (self.idPOPUP_SORTBYSIZE, _("Sort by Size"))
		sort_menu.Append (self.idPOPUP_SORTBYDATE, _("Sort by Date"))
		sort_menu.AppendSeparator ()
		sort_menu.AppendCheckItem (self.idPOPUP_ORDER, _("Order Directories-Links-Files"))
		menu.AppendMenu (wxNewId (), _("Sort Order"), sort_menu)

		if self.directories_first:
			menu_item = menu.FindItemById (self.idPOPUP_ORDER)
			menu_item.Check (true)

		EVT_MENU (self, self.idPOPUP_DOWNLOAD, self.onDownload)
		EVT_MENU (self, self.idPOPUP_RENAME, self.onRename)
		EVT_MENU (self, self.idPOPUP_DELETE, self.onDelete)
		EVT_MENU (self, self.idPOPUP_CREATE, self.onCreate)
		EVT_MENU (self, self.idPOPUP_REMOVE, self.onRemove)
		EVT_MENU (self, self.idPOPUP_CHANGE, self.onChange)
		EVT_MENU (self, self.idPOPUP_CHMOD, self.onChmod)
		EVT_MENU (self, self.idPOPUP_REFRESH, self.onRefresh)
		EVT_MENU (self, self.idPOPUP_UNSORT, self.unsorted)
		EVT_MENU (self, self.idPOPUP_SORTBYNAME, self.sortByName)
		EVT_MENU (self, self.idPOPUP_SORTBYSIZE, self.sortBySize)
		EVT_MENU (self, self.idPOPUP_SORTBYDATE, self.sortByDate)
		EVT_MENU (self, self.idPOPUP_ORDER, self.toggleDirectoriesFirst)

		return menu

	def updateStatusBar (self, dir):
		AbstractFileWindow.updateStatusBar (self, dir)
		event = events.wxDirChangeEvent (dir)
		main.getApp ().postEvent (event)

	def onRemoteWindowEvent (self, event):
		if event.kind == events.wxRemoteWindowEvent.EVT_LIST_UPDATE:
			listing = event.msg
			self.displayFreshListing (listing)
		elif event.kind == events.wxRemoteWindowEvent.EVT_LIST_REFRESH:
			if event.msg == self.getDir ():
				self.onRefresh (event)
		elif event.kind == events.wxRemoteWindowEvent.EVT_CWD:
			self.updateStatusBar (event.msg)
			self.setDir (event.msg)

	def onDownload (self, event):
		selected = util.getSelected (self.list)
		main_thread = main.getApp ().getThreadManager ().getMainThread ()

		if not main_thread:
			return

		opts = main_thread.getOpts ()
		for item in selected:
			file = util.getColumnText (self.list, item, 0)
			if file == '..':
				continue
			flags = util.getColumnText (self.list, item, 3)

			listdata = self.cache.findNode (self.getDir ())
			# Extract the true size
			for x in listdata.data:
				if x[0] == file:
					truesize = long (x[4])
					break

			event = events.wxQueueEvent (
				file        = file,
				host        = opts['host'],
				port        = opts['port'],
				username    = opts['username'],
				password    = opts['password'],
				size        = truesize,
				flags       = flags[0],
				remote_path = self.getDir (),
				local_path  = main.getApp ().getMainWindow ().getLocalWindow ().getDir (),
				direction   = protocol.DOWNLOAD,
				attempt     = 1,
				maxattempts = opts['retries'],
				method      = main.getApp ().getTransferMethod (),
			)
			main.getApp ().postEvent (event)

	def onRename (self, event):
		selected = util.getSelected (self.list)
		thread = self.getMainThread ()
		for item in selected:
			file = util.getColumnText (self.list, item, 0)
			if file == '..':
				continue
			new_name = messages.displayInputDialog (self, _("Ftpcube - Rename"),
				_("Rename %(f)s to:") %{ 'f' : file })
			thread.rename (file, new_name)
		thread.list ()

	def onDelete (self, event):
		selected = util.getSelected (self.list)
		thread = self.getMainThread ()
		for item in selected:
			file = util.getColumnText (self.list, item, 0)
			if file == '..':
				continue
			type = util.getColumnText (self.list, item, 3)
			if type[0] in '-lf':
				thread.delete (file)
			else:
				thread.rmdir (file)
		thread.list ()

	def onCreate (self, event):
		thread = self.getMainThread ()
		dir = messages.displayInputDialog (self,
			_("Ftpcube - Create Remote Directory"),
			_("Directory Name:"))
		if dir and not dir == '..':
			thread.mkdir (dir)
			thread.list ()

	def onRemove (self, event):
		selected = util.getSelected (self.list)
		thread = self.getMainThread ()
		for item in selected:
			dir = util.getColumnText (self.list, item, 0)
			if dir == '..':
				continue
			type = util.getColumnText (self.list, item, 3)
			if type[0] == 'd':
				thread.rmdir (dir)
		thread.list ()

	def onChange (self, event):
		remote_browser = browser.RemoteBrowser (self, self.cache, self.getDir ())
		ret = remote_browser.ShowModal ()
		if ret == wxID_OK:
			dir = remote_browser.getDirectory ()
			if dir:
				self.changeDirectory (dir, self.getDir ())

	def onChmod (self, event):
		selected = util.getSelected (self.list)
		thread = self.getMainThread ()
		for item in selected:
			file = util.getColumnText (self.list, item, 0)
			if file == '..':
				continue
			current_perm = self.getFilePermission (item)
			chmod_win = chmodwin.ChmodWindow (self, file, current_perm)
			ret = chmod_win.ShowModal ()
			if ret == wxID_OK:
				perm = chmod_win.getPermissions ()
				perm_str = oct (perm)
				thread.chmod (file, perm_str)

	def getFilePermission (self, selected):
		flags = util.getColumnText (self.list, selected, 3)
		if not flags:
			return None
		# Grab the permissions part
		flags = flags[1:]

		mode = 0
		if flags[0] == 'r':
			mode |= (4 << 6)
		if flags[1] == 'w':
			mode |= (2 << 6)
		if flags[2] == 'x':
			mode |= (1 << 6)
		if flags[3] == 'r':
			mode |= (4 << 3)
		if flags[4] == 'w':
			mode |= (2 << 3)
		if flags[5] == 'x':
			mode |= (1 << 3)
		if flags[6] == 'r':
			mode |= 4
		if flags[7] == 'w':
			mode |= 2
		if flags[8] == 'x':
			mode |= 1
		return mode

	def onRefresh (self, event):
		self.cache.removeNode (self.getDir ())
		thread = self.getMainThread ()
		thread.list ()

	def clearCache (self):
		del self.cache
		self.cache = cache.CacheTree ()

	def getTrueFileSize (self, file):
		listing = self.getListing ()
		if listing:
			for entry in listing:
				entry = self.parseListEntry (entry)
				if entry is not None and entry[0] == file:
					return entry[4]
		return None

	def getListing (self):
		path_node = self.cache.findNode (self.getDir())
		if path_node is None or not path_node.data:
			thread = self.getMainThread ()
			thread.list ()
		else:
			self.setList (path_node.data)

	def setList (self, list):
		if self.sorted:
			list = self.sortList (list)
		self.list.Freeze ()
		self.clearList ()
		for i in range (len (list)):
			item = list[i]
			img_index = self.selectBitmapIndex (item[3])
			self.list.InsertImageStringItem (i, item[0], img_index)
			self.list.SetStringItem (i, 1, item[1])
			self.list.SetStringItem (i, 2, item[2])
			self.list.SetStringItem (i, 3, item[3])
			self.list.SetItemData (i, i)
		self.list.Thaw ()

	def getList (self):
		list = [ ]
		count = list.GetItemCount ()
		for i in range (count):
			entry = [
				util.getColumnText (self.list, i, 0),
				util.getColumnText (self.list, i, 1),
				util.getColumnText (self.list, i, 2),
				util.getColumnText (self.list, i, 3),
			]
			list.append (tuple (entry))
		return list

	def updateListing(self, dir):
		if dir == '.':
			pass
		else:
			if dir == '..':
				dir = self.join([ dir, self.getDir () ])
			self.changeDirectory (dir, self.getDir ())
		self.getListing()

	def displayFreshListing (self, listing):
		listitems = [ ]
		cache_node = self.cache.addNode (self.getDir ())

		for entry in listing:
			parsed = self.parseListEntry (entry)
			if parsed is None:
				continue
			file, truesize, date, flags = parsed

			# Beautify the size string
			size = util.beautifySize (truesize)

			# Check if the file is a line and if so, clean up the file name
			if flags and flags[0] == 'l':
				match = self.link_re.match (file)
				if match is not None:
					file, dest = match.groups ()
				else:
					# Try the broken link match
					match = self.broken_link_re.match (file)
					if match is not None:
						file = match.group (1)
						# Extract the file
						index = file.rfind ('/') + 1
						file = file[index:]
					else:
						if __debug__:
							print "WARNING: Match failed for: %s" %file

			listitems.append ((file, size, date, flags, str (truesize)))

		# Check if the server provided us with a '..' entry.
		# Otherwise, make one.
		found = [ i for i in listitems if i[0] == '..' ]
		if not found:
			match = self.time_re.match (time.ctime (time.time ()))
			localtime = match.group (1)
			dotdotentry = ( '..', _("0 Bytes"), localtime, '', '0')
			listitems.insert (0, dotdotentry)

		# Keep the listing around in case we need it
		cache_node.data = listitems
		self.setList (listitems)

	def parseListEntry (self, entry):
		# Try the usual list matching
		match = self.list_re.match (entry)
		if match:
			flags, hardlinks, user, group, size, date, file =  match.groups ()
			return file, long (size), date, flags

		# Try the MS ftp list matching
		match = self.ms_list_re.match(entry)
		if match:
			flags, group, user, size, date, file = match.groups ()
			return file, long (size), date, flags

		if __debug__:
			print "WARNING: listing match failed for: [%s]" %entry
		return None

	def sortList (self, list):
		if self.directories_first:
			dirs = [ i for i in list if not i[3] or i[3][0] == 'd' ]
			links = [ i for i in list if i[3] and i[3][0] == 'l' ]
			files = [ i for i in list if i[3] and i[3][0] in 'f-' ]

			sorted = [ ]
			sorted.extend (self.performSort (dirs))
			sorted.extend (self.performSort (links))
			sorted.extend (self.performSort (files))
			return sorted
		return self.performSort (list)

class CwdErrorHandler (dispatcher.ErrorHandler):

	def __init__ (self, dir, remote_win, logger):
		dispatcher.ErrorHandler.__init__ (self)
		self.dir = dir
		self.remote_win = remote_win
		self.logger = logger

	def handle (self, error, dispatcher):
		if self.logger:
			self.logger.log ('error', error)
		self.remote_win.updateStatusBar (self.dir)
		self.remote_win.setDir (self.dir)
