#
# FtpCube
# Copyright (C) 2002 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import dispatcher
import protocol

import socket
import select
import errno
import time
import ftplib
import os.path
import exceptions

CRLF = '\r\n'

class FtpException (exceptions.Exception):

	def __init__ (self, args=None):
		self.args = args

class FTPDispatcher (dispatcher.Dispatcher):

	def __init__ (self, **kwargs):
		dispatcher.Dispatcher.__init__ (self, **kwargs)

		# Initialize instance variables
		self.sock = None
		self.file = None
		self.multiline_code = None
		self.incoming = [ ]

		try:
			self.host = kwargs['host']
		except KeyError:
			self.host = 'localhost'

		try:
			self.port = kwargs['port']
		except KeyError:
			self.port = 21

	def connect (self):
		# Create the ftp socket
		proto = socket.getprotobyname ('tcp')
		try:
			self.sock = socket.socket (socket.AF_INET, socket.SOCK_STREAM, proto)
		except socket.error, strerror:
			self.closeSocket ()
			raise FtpException (_("Error opening ftp socket: %(err)s")
			                      %{ 'err' : strerror })

		self.log ('local', _("Connecting to %(h)s on port %(p)d."
		                   %{ 'h' : self.host, 'p' : self.port }))

		self.sock.setblocking (0)

		try:
			err = self.sock.connect_ex ((self.host, self.port))
		except socket.error, strerror:
			self.closeSocket ()
			raise FtpException (_("Error connecting to host %(h)s: %(err)s")
			                      %{ 'h' : self.host, 'err' : strerror})

		self.log ('local', _("Connected to %(h)s." %{ 'h' : self.host}))

		while (err == errno.EINPROGRESS or err == errno.EWOULDBLOCK) and \
		      not self.die:
			time.sleep (0.25)
			err = self.sock.getsockopt (socket.SOL_SOCKET, socket.SO_ERROR)

		if err != 0 or self.die:
			self.closeSocket ()
			raise FtpException (_("Error connecting to host %(h)s: %(err)s")
			                      %{ 'h' : self.host, 'err' : os.strerror (err)})
		else:
			self.sock.setblocking (1)
			self.file = self.sock.makefile ('rb')

	def getHost (self):
		return self.host

	def getPort (self):
		return self.port

	def closeSocket (self):
		if self.sock is not None:
			self.sock.close ()
			self.sock = None
		if self.file is not None:
			self.file.close ()
			self.file = None

	def removeEndLine (self, line):
		if line[-2:] == CRLF:
			line = line[:-2]
		elif line[-1:] in CRLF:
			line = line[:-1]
		return line

	def waitFtpResponse (self):
		while 1:
			line = self.file.readline ()

			if not line:
				self.log ('local', _("Connection to %(h)s terminated...")
				                   %{ 'h' : self.host })
				self.closeSocket ()
				return

			line = self.removeEndLine (line)
			data = self.processData (line)
			if data is not None:
				return data

	def waitFtpResponseWithTimeout (self, timeout):
		read, write, ex = select.select ([ self.sock ], [ ], [ ], timeout)
		if self.sock in read:
			self.waitFtpResponse ()

	def processData (self, line):
		self.log ('remote', line)
		self.incoming.append (line)

		if line[3:4] == '-' or self.multiline_code is not None:
			if self.multiline_code is None:
				self.multiline_code = line[:3]
				return None
			elif line[:3] == self.multiline_code and line[3:4] != '-':
				self.multiline_code = None
				data = self.incoming
				self.incoming = [ ]
				return data
		else:
			data = self.incoming
			self.incoming = [ ]
			return data

	def send (self, msg, alternate=None):
		if alternate is not None:
			self.log ('local', alternate)
		else:
			self.log ('local', msg)
		return self.sock.sendall (msg + CRLF)

	def sendOOB (self, msg):
		sent = self.sock.sendall (msg + CRLF, socket.MSG_OOB)
		# Log after the send in case we throw an exception
		self.log ('local', msg)
		return sent

	def isConnected (self):
		return self.sock is not None

	def destroy (self):
		dispatcher.Dispatcher.destroy (self)
		self.closeSocket ()

	def abort (self):
		if self.isConnected ():
			self.sendOOB ('ABOR')
			response = self.waitFtpResponse ()
			if response[-1][:3] not in ('426', '226'):
				raise FtpException (response[-1])

class FTPAbstractCmd (dispatcher.DispatcherCmd):
	pass

class FTPConnect (FTPAbstractCmd):

	def execute (self, dispatcher):
		dispatcher.connect ()

class FTPAbortCmd (FTPAbstractCmd):

	def execute (self, dispatcher):
		dispatcher.sendOOB ('ABOR')
		self.response = dispatcher.waitFtpResponse ()
		if self.response[-1][:3] not in ('426', '226'):
			raise FtpException (self.response[-1])
		return self.response

class FTPGetWelcomeMsg (FTPAbstractCmd):

	def execute (self, dispatcher):
		response = dispatcher.waitFtpResponse ()
		c = response[-1][:1]
		if c not in '123':
			raise FtpException (response[-1])

class FTPCmd (FTPAbstractCmd):

	def __init__ (self, cmd):
		FTPAbstractCmd.__init__ (self)
		self.cmd = cmd

	def execute (self, dispatcher):
		dispatcher.send (self.cmd)
		self.response = dispatcher.waitFtpResponse ()
		if self.response[-1][0] != '2':
			raise FtpException (self.response[-1])
		return self.response

class FTPResponseCmd (FTPCmd):

	def __init__ (self, cmd):
		FTPCmd.__init__ (self, cmd)
		self.response = None

	def getResponse (self):
		# Set in FTPCmd.execute ()
		return self.response

class FTP257Cmd (FTPResponseCmd):

	def __init__ (self, cmd, status=None):
		FTPResponseCmd.__init__ (self, cmd)
		if status and not isinstance (status, dispatcher.DispatchStatus):
			raise FtpException (_("Invalid status object"))
		self.status = status

	def execute (self, dispatcher):
		dispatcher.send (self.cmd)
		self.response = dispatcher.waitFtpResponse ()
		self.response = self.response[-1]
		if self.response[:3] != '257':
			raise FtpException (_("Invalid 257 response: %(res)s")
			                      %{ 'res' : self.response })
		if self.status:
			dir = ftplib.parse257 (self.response)
			self.status.setOptionalData (dir)
			self.status.finished ()

class FTPPendingCmd (FTPAbstractCmd):

	def __init__ (self, cmd):
		FTPAbstractCmd.__init__ (self)
		self.cmd = cmd

	def execute (self, dispatcher):
		dispatcher.send (self.cmd)
		self.response = dispatcher.waitFtpResponse ()
		if self.response[-1][0] != '3':
			raise FtpException (self.response[-1])

class FTPLogin (FTPAbstractCmd):

	def __init__ (self, user, pw, acct):
		FTPAbstractCmd.__init__ (self)
		self.user = user
		self.pw = pw
		self.acct = acct

	def execute (self, dispatcher):
		usr_string = "USER " + self.user
		dispatcher.send (usr_string)
		response = dispatcher.waitFtpResponse ()

		if response[-1][0] == '3':
			pw_string = "PASS " + self.pw
			# So we don't print the password to screen
			alternate = "PASS " + '*' * len (self.pw)
			dispatcher.send (pw_string, alternate)
			response = dispatcher.waitFtpResponse ()

		if response[-1][0] == '3':
			# For now, assume ACCT to be ''
			acct_string = "ACCT "
			dispatcher.send (acct_string)
			response = dispatcher.waitFtpResponse ()

		if response[-1][0] != '2':
			raise FtpException (_("Error logging in to %(h)s: %(res)s")
			                      %{ 'h' : dispatcher.getHost (), 'res' : response[-1]})

class FTPTransferConnection:

	def __init__ (self, host, port, dispatcher=None, passive=True):
		self.abort_flag = False
		self.host = host
		self.port = port
		self.dispatcher = dispatcher

		if passive:
			proto = socket.getprotobyname ('tcp')
			try:
				self.sock = socket.socket (socket.AF_INET, socket.SOCK_STREAM, proto)
			except socket.error, strerror:
				raise FtpException (_("Error creating socket for transfer: %(err)s")
				                      %{ 'err' : strerror })

			self.sock.setblocking (0)

			try:
				err = self.sock.connect_ex ((host, port))
			except socket.error, strerror:
				raise FtpException (_("Error creating socket for transfer: %(err)s")
				                      %{ 'err' : strerror })

			while (err == errno.EINPROGRESS or err == errno.EWOULDBLOCK) and \
			      not self.abort_flag:
				time.sleep (0.25)
				err = self.sock.getsockopt (socket.SOL_SOCKET, socket.SO_ERROR)

			if err != 0 or self.abort_flag:
				raise FtpException (_("Error creating socket for transfer: %(err)s")
				                      %{ 'err' : os.strerror (err) })
			else:
				self.sock.setblocking (1)
		else:
			try:
				self.sock = socket.socket (socket.AF_INET, socket.SOCK_STREAM)
				self.sock.bind (('', 0))
				self.sock.listen (1)
				port = self.sock.getsockname ()[1]
				host = socket.gethostbyname (dispatcher.getHost ())
			except socket.error, (strerror):
				if self.sock:
					self.sock.close ()
				raise FtpException (_("Error creating socket for transfer: %(err)s")
				                      %{ 'err' : strerror })
			self.sendPort (host, port)

		self.file = self.sock.makefile ('rb')

	def sendPort (self, host, port):
		hbytes = host.split ('.')
		pbytes = [ str (port / 256), str (port % 256) ]
		bytes = hbytes + pbytes

		class Handler (dispatcher.ErrorHandler):
			def __init__ (self, transfer):
				self.transfer = transfer
			def handle (self, error, dispatcher):
				dispatcher.log ('error', error)
				self.transfer.abort ()

		cmd = FTPResponseCmd ('PORT ' + ','.join (bytes))
		err_handler = cmd.getErrorHandler ()
		cmd.setErrorHandler (Handler (self))
		cmd.execute (self.dispatcher)
		cmd.setErrorHandler (err_handler)

	def finalizeNonPassiveTransfer (self):
		new_sock, sockaddr = self.sock.accept ()
		self.sock.close ()
		self.file.close ()
		self.sock = new_sock
		self.file = self.sock.makefile ('rb')

	def getSock (self):
		return self.sock

	def getFile (self):
		return self.file

	def close (self):
		self.file.close ()
		self.sock.close ()

	def abort (self):
		self.abort_flag = True

class FTPTransferCmd (FTPAbstractCmd):

	def __init__ (self, cmd):
		FTPAbstractCmd.__init__ (self)
		self.cmd = cmd
		self.size = None

	def execute (self, dispatcher):
		dispatcher.send (self.cmd)
		response = dispatcher.waitFtpResponse ()
		if response[-1][0] != '1':
			raise FtpException (_("Error initiating transfer: %(res)s")
			                      %{ 'res' : response[-1]})
		if response[-1][:3] == '150':
			self.size = ftplib.parse150 (response[-1])

	def getSize (self):
		return self.size

class FTPTransfer (FTPAbstractCmd):

	def __init__ (self, cmd, rest=None, passive=True):
		FTPAbstractCmd.__init__ (self)
		self.cmd = cmd
		self.rest = rest
		self.passive = passive
		self.transfer = None

	def execute (self, dispatcher):
		if self.passive:
			cmd = FTPResponseCmd ('PASV')
			cmd.execute (dispatcher)
			host, port = ftplib.parse227 (cmd.getResponse ()[-1])
			self.transfer = FTPTransferConnection (host, port)

			if self.rest is not None:
				cmd = FTPPendingCmd ("REST %s" %self.rest)
				cmd.execute (dispatcher)

			cmd = FTPTransferCmd (self.cmd)
			cmd.execute (dispatcher)
			self.size = cmd.getSize ()
		else:
			self.transfer = FTPTransferConnection ('localhost', 0, dispatcher, False)

			if self.rest is not None:
				cmd = FTPPendingCmd ("REST %s" %self.rest)
				cmd.execute (dispatcher)

			cmd = FTPTransferCmd (self.cmd)
			cmd.execute (dispatcher)
			self.size = cmd.getSize ()

			# Finalize the non-passive connection
			self.transfer.finalizeNonPassiveTransfer ()

	def abort (self):
		self.transfer.abort ()

class FTPList (FTPTransfer):

	def __init__ (self, cmd, status=None, passive=True):
		FTPTransfer.__init__ (self, cmd, passive=passive)

		if status and not isinstance (status, dispatcher.DispatchStatus):
			raise FtpException (_("Invalid status object"))
		self.status = status

		self.abort_flag = False
		self.data = [ ]  # Local copy of data

	def execute (self, dispatcher):
		# Set the appropriate transfer type
		type = FTPResponseCmd ('TYPE A')
		type.execute (dispatcher)

		# Now prepare for the transfer
		FTPTransfer.execute (self, dispatcher)

		self.sock = self.transfer.getSock ()
		self.file = self.transfer.getFile ()

		if self.status:
			self.status.start ()

		while not self.abort_flag:
			line = self.file.readline ()
			if not line:
				break
			if line[-2:] == CRLF:
				line = line[:-2]
			elif line[-1:] in CRLF:
				line = line[:-1]

			if self.status:
				self.status.update (len (line))

			self.data.append (line)

		self.transfer.close ()
		response = dispatcher.waitFtpResponse ()
		if self.status:
			self.status.setOptionalData (self.data)
			self.status.finished ()

	def getData (self):
		return self.data

	def abort (self):
		self.transfer.abort ()
		self.abort_flag = True

class FTPAsciiTransfer (FTPTransfer):

	def __init__ (self, cmd, local_file, download=True,
	                    status=None, passive=True):
		FTPTransfer.__init__ (self, cmd, passive=passive)
		self.local_file = local_file

		if not isinstance (status, dispatcher.DispatchStatus):
			raise FtpException (_("Invalid status object"))
		self.status = status

		self.download = download
		self.abort_flag = False

	def execute (self, dispatcher):
		type = FTPResponseCmd ('TYPE A')
		type.execute (dispatcher)

		FTPTransfer.execute (self, dispatcher)

		self.sock = self.transfer.getSock ()
		self.file = self.transfer.getFile ()

		if self.status:
			self.status.start ()

		if self.download:
			local_file = open (self.local_file, 'w')
			while not self.abort_flag:
				line = self.file.readline ()
				if not line:
					break
				if self.status:
					self.status.update (len (line))
				local_file.write (line)
		else:
			local_file = open (self.local_file, 'r')
			while not self.abort_flag:
				line = local_file.readline ()
				if not line:
					break
				self.sock.sendall (line)
				if self.status:
					self.status.update (len (line))

		self.transfer.close ()
		dispatcher.waitFtpResponse ()
		if self.status:
			self.status.finished ()

	def abort (self):
		self.transfer_abort ()
		self.abort_flag = True

class FTPBinaryTransfer (FTPTransfer):

	def __init__ (self, cmd, local_file, download=True, status=None,
	                    blocksize=8192, rest=None, passive=True):
		FTPTransfer.__init__ (self, cmd, rest=rest, passive=passive)

		if not isinstance (status, dispatcher.DispatchStatus):
			raise FtpException (_("Invalid status object"))
		self.status = status
		self.local_file = local_file
		self.blocksize = blocksize
		self.download = download

		self.abort_flag = False

	def execute (self, dispatcher):
		type = FTPResponseCmd ('TYPE I')
		type.execute (dispatcher)

		FTPTransfer.execute (self, dispatcher)

		self.sock = self.transfer.getSock ()

		if self.status:
			self.status.start ()

		if self.download:
			if self.rest is not None:
				local_file = open (self.local_file, 'ab')
			else:
				local_file = open (self.local_file, 'wb')

			while not self.abort_flag:
				data = self.sock.recv (self.blocksize)
				if not data:
					break
				if self.status:
					self.status.update (len (data))
				local_file.write (data)
		else:
			local_file = open (self.local_file, 'rb')

			if self.rest is not None:
				local_file.seek (self.rest)

			while not self.abort_flag:
				data = local_file.read (self.blocksize)
				if not data:
					break
				self.sock.sendall (data)
				if self.status:
					self.status.update (len (data))

		local_file.close ()
		self.transfer.close ()
		dispatcher.waitFtpResponse ()
		if self.status:
			self.status.finished ()

	def abort (self):
		self.transfer.abort ()
		self.abort_flag = True

class FTP (protocol.ProtocolInterface):

	def __init__ (self, **kwargs):

		# Grab our crucial arguments
		try:
			self.host = kwargs['host']
		except KeyError:
			raise FtpException (_("Error opening FTP connection: No host given"))
		try:
			self.port = kwargs['port']
		except KeyError:
			raise FtpException (_("Error opening FTP connection: Invalid port given"))
		try:
			self.logger = kwargs['logger']
		except KeyError:
			self.logger = None

		# Initialize instance variables
		self.passive = True

		# Create the dispatcher
		self.dispatcher = FTPDispatcher (**kwargs)
		self.dispatcher.start ()

	def setPassive (self, passive):
		self.passive = passive

	def getPassive (self):
		return self.passive

	def getDispatcher (self):
		return self.dispatcher

	def destroy (self):
		# Depending on our state, we may or may not be able
		# to destroy the FTP connection object gracefully.
		try:
			self.quit ()
		except Exception, strerror:
			pass
		self.dispatcher.destroy ()

	def busy (self):
		return self.dispatcher.busy ()

	def addOutgoing (self, cmd, handler):
		if handler is None:
			if self.logger is not None:
				handler = dispatcher.ErrorLogger (self.logger)
				cmd.setErrorHandler (handler)
		else:
			cmd.setErrorHandler (handler)
		self.dispatcher.dispatch (cmd)

	def initiateConnect (self, handler=None):
		connect = FTPConnect ()
		self.addOutgoing (connect, handler)

	def getWelcomeMsg (self, handler=None):
		welcome = FTPGetWelcomeMsg ()
		self.addOutgoing (welcome, handler)

	def login (self, username, password, handler=None):
		login = FTPLogin (username, password, '')
		self.addOutgoing (login, handler)

	def noop (self, handler=None):
		ftpcmd = FTPCmd ('NOOP')
		self.addOutgoing (ftpcmd, handler)

	def cwd (self, dir, handler=None):
		if not isinstance (dir, str) or dir == '.':
			return
		elif dir == '..':
			cmd = 'CDUP'
		else:
			cmd = 'CWD ' + dir
		ftpcmd = FTPCmd (cmd)
		self.addOutgoing (ftpcmd, handler)

	def pwd (self, status=None, handler=None):
		cmd = 'PWD'
		ftpcmd = FTP257Cmd (cmd, status=status)
		self.addOutgoing (ftpcmd, handler)

	def delete (self, file, handler=None):
		cmd = 'DELE ' + file
		ftpcmd = FTPCmd (cmd)
		self.addOutgoing (ftpcmd, handler)

	def rename (self, name, new_name, handler=None):
		cmd = 'RNFR ' + name
		ftpcmd = FTPPendingCmd (cmd)
		self.addOutgoing (ftpcmd, handler)
		cmd = 'RNTO ' + new_name
		ftpcmd = FTPCmd (cmd)
		self.addOutgoing (ftpcmd, handler)

	def mkdir (self, dir, status=None, handler=None):
		cmd = 'MKD ' + dir
		ftpcmd = FTP257Cmd (cmd, status=status)
		self.addOutgoing (ftpcmd, handler)

	def rmdir (self, dir, handler=None):
		cmd = 'RMD ' + dir
		ftpcmd = FTPCmd (cmd)
		self.addOutgoing (ftpcmd, handler)

	def chmod (self, mode, file, handler=None):
		cmd = "SITE CHMOD %s %s" %(mode, file)
		ftpcmd = FTPCmd (cmd)
		self.addOutgoing (ftpcmd, handler)

	def quit (self, handler=None):
		# Perform quit inline, so we don't have to worry
		# about the connection closing
		self.dispatcher.sendOOB ('QUIT')
		self.dispatcher.waitFtpResponseWithTimeout (1.0)

	def abort (self, handler=None):
		executing = self.dispatcher.getExecutingCmd ()
		if isinstance (executing, FTPTransfer):
			executing.abort ()
			self.dispatcher.resetQueue ()
			self.addOutgoing (FTPAbortCmd (), handler)
		elif isinstance (executing, FTPConnect):
			self.dispatcher.destroy ()

	def list (self, args=None, status=None, passive=True, handler=None):
		cmd = 'LIST'
		if args is not None:
			for i in args:
				if i:
					cmd = cmd + (' ' + i)
		listcmd = FTPList (cmd, status=status, passive=passive)
		self.addOutgoing (listcmd, handler)

	def retrieveAscii (self, file, local_dir, status=None, passive=True, handler=None):
		cmd = 'RETR ' + file
		local_file = os.path.join (local_dir, file)
		asciicmd = FTPAsciiTransfer (cmd, local_file, status=status, passive=passive)
		self.addOutgoing (asciicmd, handler)

	def uploadAscii (self, file, local_dir, status=None, passive=True, handler=None):
		cmd = 'STOR ' + file
		local_file = os.path.join (local_dir, file)
		asciicmd = FTPAsciiTransfer (cmd, local_file, download=False, status=status, passive=passive)
		self.addOutgoing (asciicmd, handler)

	def retrieveBinary (self, file, local_dir, status=None, blocksize=8192, rest=None,
	                                           passive=True, handler=None):
		cmd = 'RETR ' + file
		local_file = os.path.join (local_dir, file)
		binarycmd = FTPBinaryTransfer (cmd, local_file, status=status, blocksize=blocksize,
		                                                rest=rest, passive=passive)
		self.addOutgoing (binarycmd, handler)

	def uploadBinary (self, file, local_dir, status=None, blocksize=8192, rest=None,
	                                         passive=True, handler=None):
		cmd = 'STOR ' + file
		local_file = os.path.join (local_dir, file)
		binarycmd = FTPBinaryTransfer (cmd, local_file, download=False, status=status,
		                                                blocksize=blocksize, rest=rest, passive=passive)
		self.addOutgoing (binarycmd, handler)
