#
# FtpCube
# Copyright (C) 2002 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import config
import messages
import mainwin
import threads
import logger
import icons.ftpcube

import sys
import getopt
import os
import gettext
from wxPython.wx import *

# The main application instance
app = None

class App (wxApp):

	def OnInit (self):

		# Initialize i18n
		gettext.install ('ftpcube')

		# Set the data prefixes
		self.setDefaultConfigDir ()

		self.setConfig (None)
		config_file = os.path.join (self.getConfigDir (), "config")
		try:
			configuration = config.Configuration (config_file)
			self.setConfig (configuration)
		except config.ConfigError, strerror:
			messages.displayErrorDialog (NULL, strerror)

		try:
			opts = self.parseCmdLineOpts ()
			self.setOpts (opts)
		except Exception, strerror:
			print _("Error parsing options: %(str)s") %{ 'str' : strerror }
			self.usage ()
			sys.exit (1)

		# Create the event notification registry
		self.createEventRegistry ()

		# Create the application thread manager
		self.thread_mgr = threads.ThreadManager ()

		# Initialize all wxWindows image handlers
		wxInitAllImageHandlers ()

		self.SetAppName (_("Ftpcube"))
		self.main_window = None
		return true

	def createMainWindow (self):
		self.main_window = mainwin.MainWindow (NULL, _("Ftpcube"))
		self.main_window.Show (true)
		self.SetTopWindow (self.main_window)

	def getMainWindow (self):
		return self.main_window

	def getTransferMethod (self):
		return self.config['transfer_type']

	def setOpts (self, opts):
		self.opt_dict = opts

	def getOpts (self):
		return self.opt_dict

	def setConfig (self, config):
		self.config = config

	def getConfig (self):
		return self.config

	def getThreadManager (self):
		return self.thread_mgr

	def getAppIcon (self):
		icon =  wxIcon ('ftpcube_XPM', wxBITMAP_TYPE_XPM)
		icon.CopyFromBitmap (icons.ftpcube.getBitmap ())
		return icon

	def setConfigDir (self, dir):
		self.config_dir = dir

	def getConfigDir (self):
		return self.config_dir

	def setDefaultConfigDir (self):
		if os.environ.has_key ('HOME'):
			self.setConfigDir (os.path.join (os.environ['HOME'], ".ftpcube"))
		elif os.environ.has_key ('HOMEPATH'):
			self.setConfigDir (os.path.join (os.environ['USERPROFILE'], "FtpCube Config"))
		else:
			if sys.platform == "win32":
				self.setConfigDir ('.')
			elif os.name == 'posix':
				if os.path.exists ('/usr/local/share/ftpcube'):
					self.setConfigDir ('/usr/local/share/ftpcube')
				elif os.path.exists ('/usr/share/ftpcube'):
					self.setConfigDir ('/usr/share/ftpcube')
				else:
					self.setConfigDir ('.')

	def parseCmdLineOpts (self):
		try:
			opts, args = getopt.getopt (sys.argv[1:], 'h', [
				'help',    # Print out usage info
			])
		except getopt.GetOptError:
			raise # Re-raise to alert the application

		for o, a in opts:
			if o in ('-h', '--help'):
				self.usage ()
				sys.exit (1)

	def usage (self):
		pass

	def createEventRegistry (self):
		self.event_registry = { }

	def registerEventListener (self, event_type, listener):
		try:
			self.event_registry[event_type].append (listener)
		except KeyError:
			self.event_registry[event_type] = [ listener ]

	def postEvent (self, event):
		if not self.event_registry.has_key (event.GetEventType ()):
			return None
		listeners = self.event_registry[event.GetEventType ()]
		for listener in listeners:
			wxPostEvent (listener, event)
		return listeners

	def initiateConnection (self, opts):
		if opts is None:
			return

		# Create some logging objects
		clogger, dlogger, ulogger = self.getLoggers ()

		self.thread_mgr.newMainThread (
			name=_("MainThread"),
			opts=opts,
			logger=clogger,
			download_logger=dlogger,
			upload_logger=ulogger,
		)
		thread = self.thread_mgr.getMainThread ()
		try:
			thread.initiateConnect ()
		except threads.ThreadException, strerror:
			messages.displayErrorDialog (self,
				_("Error initiating connection: ") + strerror)

	def getLoggers (self):
		# Start with empty loggers
		clogger, dlogger, ulogger = None, None, None

		config = self.getConfig ().getOpts ()
		if config['logging']:
			conf_dir = self.getConfigDir ()
			path = os.path.join (conf_dir, config['cmd_log'])
			clogger = logger.FileAndConsoleLogger (path)
			path = os.path.join (conf_dir, config['download_log'])
			dlogger = logger.FileLogger (path)
			path = os.path.join (conf_dir, config['upload_log'])
			ulogger = logger.FileLogger (path)
		else:
			clogger = logger.ConsoleLogger ()

		return (clogger, dlogger, ulogger)


def getApp ():
	"""Returns the application instance in the global main namespace."""
	return app

def main (args):
	"""Main program routine.

	Creates a global instance of the application and sets it in the main
	module namespace."""
	global app
	app = App ()
	app.createMainWindow ()
	app.MainLoop ()
